# -*- coding: utf-8 -*-
"""
Created on Mon Feb 10 13:24:59 2020

@author: Mitchell Abrams

Consolidates additional processing as in Appendix D of the FARS Analytical
User Manual.
"""
import pandas as pd
import numpy as np

import janitor
from datetime import datetime


def get_fips(df):
    """Convert State/County codes to FIPS locations."""
    state = df['STATE'].apply(lambda x: str(x).zfill(2))
    county = df['COUNTY'].apply(lambda x: str(x).zfill(3))

    fips = state + county

    return fips.where(df['COUNTY'].between(1, 996))


def time_of_day(df):
    """Return daytime/nighttime using NHTSA convention."""
    hr = df['HOUR']
    conditions = [
        (hr.between(6, 17)),
        (hr.between(0, 5) | (hr.between(18, 24))),
        (hr == 99)]
    choices = ['Daytime', 'Nighttime', 'Unknown']
    return np.select(conditions, choices, default='ERROR')


def day_of_week(df):
    """Return weekday/weekend using NHTSA convention."""
    hr = df['HOUR']
    day = df['DAY_WEEK']

    conditions = [
        (((day == 2) & hr.between(6, 23)) | day.isin([3, 4, 5]) |
         ((day == 6) & (hr.between(0, 17) | (hr == 24)))),
        (((day == 6) & hr.between(18, 23)) | day.isin([1, 7]) |
         ((day == 2) & (hr.between(0, 5) | (hr == 24)))),
        ((day == 9) | ((hr == 99) & day.isin([2, 6])))]
    choices = ['Weekday', 'Weekend', 'Unknown']
    return np.select(conditions, choices, default='ERROR')


def collision_type(df):
    """Return manner of collision using NHTSA convention."""
    coll = df['MAN_COLL']
    yr = df['YEAR']

    conditions = [
         (coll == 0), (coll == 1), (coll == 2),
         (((coll == 4) & yr.between(1975, 2001)) |
          (coll.between(3, 6) & (yr.between(2002, 2009))) |
          ((coll == 6) & (yr >= 2010))),
         (((coll == 7) & yr.between(1975, 1977)) |
          (coll.isin([5, 6]) & yr.between(1978, 2001)) |
          (coll.isin([7, 8]) & (yr >= 2002))),
         (((coll == 3) & yr.between(1975, 2001)) |
          (coll.between(9, 11) & (yr >= 2002))),
         (((coll == 9) & yr.between(1975, 2001)) |
          (coll.isin([98, 99]) & (yr >= 2002)))]
    choices = ["Not Collision", "Rear-End", "Head-On", "Angle", "Sideswipe",
               "Other", "Unknown"]

    return np.select(conditions, choices, default='ERROR')


def trafficway(df):
    """Return relative position to trafficway using NHTSA convention."""
    road = df['REL_ROAD']
    yr = df['YEAR']

    conditions = [
        ((road == 1) | ((yr >= 2001) & (road == 11))),
        (road == 2),
        (road == 3),
        (road.isin([4, 5, 6, 7, 8, 10])),
        (road.isin([9, 98, 99]))]
    choices = ["On roadway", "Off roadway/shoulder", "Off roadway/median",
               "Off roadway/other", "Unknown"]

    return np.select(conditions, choices, default='ERROR')


def functional_class(df):
    """Return roadway function class using NHTSA convention."""
    yr = df['YEAR']
    # if before 1981, return series of pd.NA
    func = df['FUNCTION']

    conditions = [
        (func.isin([1, 11])),
        ((~(yr.between(1987, 2014)) & (func == 2)) |
         ((yr.between(1987, 2014) & (func == 12)))),
        ((~(yr.between(1987, 2014)) & (func == 3)) |
         (yr.between(1987, 2014) & (func.isin([2, 13])))),
        ((~(yr.between(1987, 2014)) & (func == 4)) |
         (yr.between(1987, 2014) & (func.isin([3, 14])))),
        ((yr.between(1981, 1986) & func.isin([5, 6, 7])) |
         (yr.between(1987, 2014) & func.isin([4, 5, 15])) |
         ((yr >= 2015) & func.isin([5, 6]))),
        ((yr.between(1981, 1986) & (func == 8)) |
         (yr.between(1987, 2014) & func.isin([6, 16])) |
         ((yr >= 2015) & (func == 7))),
        ((yr.between(1981, 1986) & (func == 9)) |
         (yr.between(1987, 2014) & func.isin([9, 19, 99])) |
         ((yr >= 2015) & func.isin([96, 98, 99])))]
    choices = ["Interstate, principal arterial",
               "Freeway and expressway, principal arterial",
               "Principal arterial, other",
               "Minor arterial",
               "Collector",
               "Local",
               "Unknown"]
    return np.select(conditions, choices, default=np.nan)


def land_use(df):
    """Return rural/urban land use using NHTSA convention."""
    yr = df['YEAR']

    func1 = df['LAND_USE']
    func2 = df['FUNC_SYS']
    if 'RUR_URB' in df.columns:
        func3 = df['RUR_URB']
    else:
        func3 = pd.Series(0, index=np.arange(len(func1)))

    conditions = [
        ((yr.between(1981, 1986) & (func1 == 2)) |
         (yr.between(1987, 2014) & ((func2 == 9) | func2.between(1, 6))) |
         ((yr >= 2015) & (func3 == 1))),
        ((yr.between(1981, 1986) & (func1 == 1)) |
         (yr.between(1987, 2014) & ((func2 == 19) | func2.between(11, 16))) |
         ((yr >= 2015) & (func3 == 2))),
        ((yr.between(1981, 1986) & (func1 == 9)) |
         (yr.between(1987, 2014) & (func2 == 99)) |
         ((yr >= 2015) & func3.isin([6, 8, 9])))]
    choices = ["Rural", "Urban", "Unknown"]
    return np.select(conditions, choices, default='ERROR')


def interstate(df):
    """Return whether road is an interstate using NHTSA convention."""
    yr = df['YEAR']
    func = df['ROUTE'].where(yr.between(1975, 1980)).add(
        df['ROAD_FNC'].where(yr.between(1981, 2014)), fill_value=0).add(
        df['FUNC_SYS'].where((yr >= 2015)), fill_value=0)

    conditions = [
        ((func == 1) | (func == 11)),
        (func.between(2, 8) | func.between(12, 16)),
        func.isin([9, 19, 96, 98, 99])]
    choices = ["Interstate", "Non Interstate", "Unknown"]
    return np.select(conditions, choices, default='ERROR')


def is_passenger_car(df):
    """Return if vehicle is a passenger car, per NHTSA convention."""
    yr = df['YEAR']
    body = df['BODY_TYP']

    return ((yr.between(1975, 1981) & body.between(1, 9)) |
            (yr.between(1982, 1990) & (body.between(1, 11) | (body == 67))) |
            ((1991 <= yr) & (body.between(1, 11) | (body == 17))))


def is_light_truck_or_van(df):
    """Return if vehicle is a light truck or van, per NHTSA convention."""
    yr = df['YEAR']
    body = df['BODY_TYP']
    tow_veh = df['TOW_VEH']

    return ((yr.between(1975, 1981) & (body.isin([43, 50, 51, 52]) |
                                       ((body == 60) & (tow_veh == 0)))) |
            (yr.between(1982, 1990) & (body.isin([12, 40, 41, 48, 49, 50, 51,
                                                  53, 54, 55, 56, 58, 59, 68,
                                                  69]) |
                                       ((body == 79) &
                                        (tow_veh.isin([0, 9]))))) |
            ((1991 <= yr) & (body.isin([14, 15, 16, 19, 20, 21, 22, 24, 25]) |
                             body.between(28, 41) | body.between(45, 49) |
                             ((body == 79) & (tow_veh.isin([0, 9]))))))


def is_large_truck(df):
    """Return if vehicle is a large truck, per NHTSA convention."""
    yr = df['YEAR']
    body = df['BODY_TYP']
    tow_veh = df['TOW_VEH']
    return ((yr.between(1975, 1981) & (body.between(53, 59) |
                                       ((body == 60) & (tow_veh == 1)))) |
            (yr.between(1982, 1990) & (body.isin([70, 71, 72, 74,
                                                  75, 76, 78]) |
                                       ((body == 79) &
                                        tow_veh.between(1, 5)))) |
            ((1991 <= yr) & (body.isin([60, 61, 62, 63, 64,
                                        66, 67, 71, 72, 78]) |
                             ((body == 79) & tow_veh.between(1, 4)))))


def is_motorcycle(df):
    """Return if vehicle is a motorcycle, per NHTSA convention."""
    yr = df['YEAR']
    body = df['BODY_TYP']

    return ((yr.between(1975, 1981) & body.between(15, 18)) |
            (yr.between(1982, 1990) & body.between(20, 29)) |
            ((1991 <= yr) & body.between(80, 89)))


def is_bus(df):
    """Return if vehicle is a bus, per NHTSA convention."""
    yr = df['YEAR']
    body = df['BODY_TYP']

    return ((yr.between(1975, 1981) & body.between(25, 29)) |
            (yr.between(1982, 1990) & body.between(30, 39)) |
            ((1991 <= yr) & body.between(50, 59)))


def is_other_or_unknown(df):
    """Return if vehicle is other/unknown, per NHTSA convention."""
    yr = df['YEAR']
    body = df['BODY_TYP']
    tow_veh = df['TOW_VEH']

    return ((yr.between(1975, 1981) & (body.between(35, 42) |
                                       body.isin([44, 45, 99]))) |
            (yr.between(1982, 1990) & (body.isin([13, 14, 42, 52, 73, 77,
                                                  80, 81, 82, 83, 88, 89,
                                                  90, 99]))) |
            ((1991 <= yr) & (body.isin([12, 13, 23, 42, 65, 73, 90,
                                        91, 92, 93, 94, 95, 96, 97,
                                        99, 98]) |
                             ((body == 79) & tow_veh.isin([5, 6])))))


def is_passenger_vehicle(df):
    """Return if vehicle is a passenger vehicle, per NHTSA convention."""
    return is_passenger_car(df) & is_light_truck_or_van(df)


def is_utility_vehicle(df):
    """Return if vehicle is a utility vehicle, per NHTSA convention."""
    yr = df['YEAR']
    body = df['BODY_TYP']

    return ((yr.between(1975, 1981) & (body == 43)) |
            (yr.between(1982, 1990) & body.isin([12, 56, 68])) |
            ((1991 <= yr) & body.isin([14, 15, 16, 19])))


def is_pickup(df):
    """Return if vehicle is a pickup truck, per NHTSA convention."""
    yr = df['YEAR']
    body = df['BODY_TYP']

    return ((yr.between(1975, 1981) & (body == 50)) |
            (yr.between(1982, 1990) & body.isin([50, 51])) |
            ((1991 <= yr) & body.between(30, 39)))


def is_van(df):
    """Return if vehicle is a van, per NHTSA convention."""
    yr = df['YEAR']
    body = df['BODY_TYP']

    return ((yr.between(1975, 1981) & (body == 51)) |
            (yr.between(1982, 1990) & body.isin([40, 41, 48, 49])) |
            ((1991 <= yr) & body.isin([20, 21, 22, 24, 25, 28, 29])))


def is_medium_truck(df):
    """Return if vehicle is a medium truck, per NHTSA convention."""
    yr = df['YEAR']
    body = df['BODY_TYP']

    return ((yr.between(1975, 1981) & body.isin([53, 54, 56])) |
            (yr.between(1982, 1990) & body.isin([70, 71, 75, 78])) |
            ((1991 <= yr) & body.isin([60, 61, 62, 64, 67, 71])))


def is_heavy_truck(df):
    """Return if vehicle is a heavy truck, per NHTSA convention."""
    yr = df['YEAR']
    body = df['BODY_TYP']
    tow_veh = df['TOW_VEH']

    return ((yr.between(1975, 1981) & (body.isin([55, 57, 58, 59]) |
                                       ((body == 60) & (tow_veh == 1)))) |
            (yr.between(1982, 1990) & (body.isin([72, 74, 76]) |
                                       ((body == 79) &
                                        tow_veh.between(1, 5)))) |
            ((1991 <= yr) & (body.isin([63, 66, 72, 78]) |
                             ((body == 79) & tow_veh.between(1, 4)))))


def is_combination_truck(df):
    """Return if vehicle is a combination truck, per NHTSA convention."""
    yr = df['YEAR']
    body = df['BODY_TYP']
    tow_veh = df['TOW_VEH']

    return ((yr.between(1975, 1981) & (body.isin([57, 58, 59]) |
                                       (body.isin([53, 54, 55, 56, 60]) &
                                        (tow_veh == 1)))) |
            (yr.between(1982, 1990) & ((body == 74) |
                                        (body.isin([70, 71, 72, 75,
                                                    76, 78, 79]) &
                                         tow_veh.between(1, 5)))) |
            ((1991 <= yr) & ((body == 66) |
                             (body.isin([60, 61, 62, 63, 64,
                                         71, 72, 78, 79]) &
                             tow_veh.between(1, 4)))))


def is_single_unit_truck(df):
    """Return if vehicle is a single unit truck, per NHTSA convention."""
    yr = df['YEAR']
    body = df['BODY_TYP']
    tow_veh = df['TOW_VEH']

    return ((yr.between(1975, 1981) & body.isin([53, 54, 55, 56, 60]) &
             (tow_veh == 0)) |
            (yr.between(1982, 1990) & body.isin([70, 71, 72, 75, 76, 78, 79]) &
             tow_veh.isin([0, 9])) |
            ((1991 <= yr) & body.isin([60, 61, 62, 63, 64,
                                       67, 71, 72, 78, 79]) &
             tow_veh.isin([0, 5, 6, 9])))


def impact_area(df, i_type):
    """Return impact area classification, per NHTSA convention."""
    yr = df['YEAR']
    if i_type == 1:
        impact = df['IMPACT1']
    else:
        impact = df['IMPACT2']

    conditions = [
        (impact == 0),
        impact.isin([1, 11, 12]),
        impact.isin([2, 3, 4, 81, 82, 83]),
        impact.isin([8, 9, 10, 61, 62, 63]),
        impact.isin([5, 6, 7]),
        impact.isin([13, 14, 15, 16, 18, 19, 20]),
        impact.isin([98, 99])]
    choices = ['Non-Collision', 'Front', 'Right Side', 'Left Side',
               'Rear', 'Other', 'Unknown']
    return np.select(conditions, choices, default='ERROR')


def license_status(df):
    """Return license status, per NHTSA convention."""
    yr = df['YEAR']
    l_status = df['L_STATUS']
    conditions = [
        ((yr.between(1975, 1981) & l_status.isin([0, 3, 7])) |
         (yr.between(1982, 1986) & l_status.isin([0, 2, 7, 8])) |
         (yr.between(1987, 1992) & l_status.isin([5, 6, 7, 8])) |
         (yr.between(1993, 2003) & l_status.isin([6, 7, 8])) |
         ((2004 <= yr) & (l_status == 6))),
        ((yr.between(1975, 1981) & l_status.isin([1, 2, 4, 5, 6])) |
         (yr.between(1982, 1986) & l_status.isin([1, 3, 4, 5, 6])) |
         ((1987 <= yr) & l_status.isin([0, 1, 2, 3, 4]))),
        ((yr.between(1975, 2010) & (l_status == 9)) |
         ((2011 <= yr) & l_status.isin([7, 9])))]
    choices = ["Valid", "Invalid", "Unknown"]
    return np.select(conditions, choices, default='ERROR')


def license_compliance(df):
    """Return license compliance, per NHTSA convention."""
    yr = df['YEAR']
    if 1981 <= yr <= 1986:
        compl = df['L_CL_VEH']
    elif 1987 <= yr:
        compl = df['L_COMPL']

    conditions = [
        ((yr.between(1982, 1986) & compl.isin([0, 2, 4])) |
         ((1987 <= yr) & compl.isin([1, 3]))),
        ((yr.between(1982, 1986) & compl.isin([1, 3, 5])) |
         ((1987 <= yr) & compl.isin([0, 2]))),
        ((yr.between(1982, 1986) & compl == 9) |
         ((1987 <= yr) & compl.isin([9, 6, 7, 8])))]
    choices = ["Valid", "Invalid", "Unknown"]
    return np.select(conditions, choices, default='ERROR')


def person_type(df):
    """Return person classification, per NHTSA convention."""
    yr = df['YEAR']
    per_typ = df['PER_TYP']

    conditions = [
        (per_typ == 1),
        (per_typ.isin([2, 9])),
        ((yr.between(1975, 1981) & per_typ == 5) |
         ((yr <= 1982) & per_typ.isin([3, 4]))),
        ((yr.between(1975, 1981) & (per_typ == 3)) |
         ((yr <= 1982) & (per_typ == 5))),
        ((yr.between(1975, 1981) & (per_typ == 4)) |
         ((yr <= 1982) & per_typ.isin([6, 7]))),
        (per_typ.isin([8, 10, 19])),
        (per_typ.isin([99, 88]))]
    choices = ["Driver", "Passenger",  "Other non-occupant", "Pedestrian",
               "Pedalcyclist", "Other/Unknown non-occupant",
               "Unknown person type"]
    return np.select(conditions, choices, default='ERROR')


def restraint_use(df):
    """Return restraint usage, per NHTSA convention."""
    yr = df['YEAR']
    rest = df['REST_USE']

    conditions = [
        rest.isin([0, 5, 6, 7, 15, 16, 17, 19, 20]),
        rest.isin([1, 2, 3, 4, 8, 10, 11, 12, 97]),
        rest.isin([13, 14]),
        rest.isin([9, 29, 98, 99])]
    choices = ["Not Used", "Used", "Used Improperly", "Unknown"]
    return np.select(conditions, choices, default='ERROR')


def helmet_use(df):
    """Return helmet usage, per NHTSA convention."""
    yr = df['YEAR']
    helmet = df['REST_USE']
    if 'REST_MIS' in df.columns:
        rest_mis = df['REST_MIS']
    else:
        rest_mis = pd.Series(pd.NA, index=df.index)

    conditions = [
        ((helmet.isin([0, 1, 2, 3, 4, 6, 10, 11, 12, 13, 14, 15, 17])) |
         ((rest_mis == 1) & (helmet.isin([5, 16, 8, 19, 97])))),
        (((yr <= 2009) & (helmet.isin([5, 8]))) |
         ((yr >= 2010) & (rest_mis == 0) & (helmet.isin([5, 16, 19, 8, 97])))),
        (helmet.isin([9, 29, 98, 99]))]
    choices = ["Not Helmeted", "Helmeted", "Unknown"]
    return np.select(conditions, choices, default='ERROR')

def air_bag_deployed(df):
    """Return air bag deployment as True/False, or NA if unknown"""
    yr = df['YEAR']
    if 'AUT_REST' in df.columns:
        air_bag = df['AUT_REST'].where(yr.between(1975, 1990))
    else:
        air_bag = pd.Series()
    if 'AIR_BAG' in df.columns:
        air_bag = air_bag.add(df['AIR_BAG'].where((yr >= 1991)), fill_value=0)

    conditions = [
        (((yr <= 1997) & (air_bag == 3)) |
         ((yr >= 1998) & (air_bag.isin([1, 2, 3, 7, 8, 9])))),
        (((yr <= 1997) & (air_bag == 4)) |
         ((yr >= 1998) & (air_bag.isin([20, 28, 30, 31, 32])))),
        (((yr <= 1997) & (air_bag == 0)) |
         ((yr >= 1998) & (air_bag.isin([0, 97])))),
        (((yr <= 1997) & (air_bag == 9)) |
         ((yr >= 1998) & (air_bag.isin([99, 98]))))]
    choices = ["Deployed", "Not Deployed", "Non-Motorist", "Unknown"]
    return np.select(conditions, choices, default=np.nan)



'''
Utility scripts for analyzing the FARS dataset
Mitchell Abrams, 2019

Includes functions and data structures which may be useful when exploring the FARS dataset. Provides dicts
mapping various FARS codes to human-readable values
across multiple years.

'''
import pandas as pd
import numpy as np
import requests, json

def createVehID(df, yr):
    x = df

    x['VEH_ID'] = x.apply((lambda a: int("{0}{1}{2:03}".format(str(yr)[-2:], int(a.ST_CASE), int(a.VEH_NO)))), axis=1)
    return x.copy()


def createPerID(df, yr):
    x = df.copy()
    if yr is None:
        x['PER_ID'] = x.apply(
            (lambda a: int("{0}{1}{2:03}{3:03}".format(str(a.YEAR)[-2:], int(a.ST_CASE), int(a.VEH_NO), int(a.PER_NO)))),
            axis=1)
    else:
        x['PER_ID'] = x.apply(
            (lambda a: int("{0}{1}{2:03}{3:03}".format(str(yr)[-2:], int(a.ST_CASE), int(a.VEH_NO), int(a.PER_NO)))),
            axis=1)
    return x


def createCaseID(df, yr):
    x = df
    x['ID'] = x.apply((lambda a: int("{0}{1}".format(str(yr)[-2:], int(a.ST_CASE)))), axis=1)#, meta=(None, 'int64'))
    return x.copy()



# -*- coding: utf-8 -*-
"""
Created on Mon Feb 17 11:04:59 2020

@author: Mitchell Abrams
"""

# import duecredit
import pandas as pd
import dask.dataframe as dd
import janitor
import os

import pickle

from pathlib import Path

from thefuzz import process as fuzzyprocess

from .builder import *
from .builder import get_renaming
import fars_cleaner.extra_info as ei

from fars_cleaner.fars_utils import createPerID

from fars_cleaner import FARSFetcher


def load_pipeline(
        start_year=1975,
        end_year=2020,
        first_run=True,
        target_folder=None,
        load_from=None,
        use_dask=False,
        client=None,
        fetcher: FARSFetcher = None,
        debug=0,
        **kwargs):
    """Load pipeline to load and process FARS data between `start_year` and `end_year.`

    Loads FARS data from NHTSA database and pre-processes in an opinionated manner.
    FARS data are preprocessed according to a preset mapping file.


    Parameters
    ----------
    start_year : int, optional
        Year to start analysis. The default is 1975.
    end_year : int, optional
        Year to end analysis. The default is 2020.
    first_run : bool, optional
        Flag to determine whether to process and write-out the required files, or whether the data can be loaded from
        disk pre-processed. The default is True.
    target_folder : Path-like, optional
        Path to target folder to save out pickled data after processing. If None, defaults to the cache directory
        specified by the FARSFetcher.
    load_from : Path-like, optional
        Path to source folder to load processed data from. If `first_run` is true, this parameter is ignored.
    use_dask : bool, optional
        Flag to determine whether to parallelize using Dask. Supported, but not recommended at this time.
        The default is False.
    client : Dask Distributed Client, optional. Required if use_dask is True.
        When provided and use_dask is False, use_dask is set to True.
    fetcher : FARSFetcher, optional
        FARSFetcher instance to facilitate file download and cache management. Overrides `load_from` value.
    debug : int, optional
        Set verbosity level.
        0 = No verbose output.
        1 = Print basic progress messages.
        2 = Print messages along with progress bars (provided by tqdm). NOT implemented yet.

    Returns
    -------
    (pandas.DataFrame, pandas.DataFrame, pandas.DataFrame)
        Returns three DataFrames: vehicles, accidents, and people.

    Raises
    ------
    ValueError
        If `use_dask` is True and no client provided.
        If `fetcher` and `load_from` are both not provided.

    """

    if use_dask and client is None:
        raise ValueError("Must provide a client to use Dask.")

    if client is not None and not use_dask:
        use_dask = True

    if fetcher is None:
        if load_from is None:
            raise ValueError("Must provide one of fetcher or load_from to access data.")
        else:
            data_dir = Path(load_from)
    else:
        data_dir = fetcher.get_data_path()

    if target_folder is None:
        target_folder = data_dir

    if debug > 0:
        print("Loading mappings...")

    if fetcher:
        mapper_file = Path(fetcher.fetch_mappers())
    else:
        mapper_file = data_dir / "mapping.dict"

    if mapper_file.exists():
        with open(mapper_file, 'rb') as f:
            mappers = pickle.load(f)
    else:
        raise FileNotFoundError("Mapping dictionary not found.")

    if debug > 0:
        print("Mappings loaded.")
        print("Loading data...")

    if fetcher:
        fetcher.fetch_subset(start_year, end_year)
        if debug > 0:
            print("Fetched originals, loading to pandas....")

    if first_run:

        lazy_people = []
        lazy_vehicles = []
        lazy_accidents = []

        for year in range(start_year, end_year + 1):
            if debug > 1:
                print(year)
            vehicle, person, accident = load_basic(year,
                                                   use_dask=use_dask,
                                                   data_dir=data_dir,
                                                   mapping=mappers,
                                                   client=client)
            accident['YEAR'] = year
            vehicle['YEAR'] = year
            person['YEAR'] = year

            lazy_people.append(person)
            lazy_vehicles.append(vehicle)
            lazy_accidents.append(accident)
            if debug > 1:
                #print('dense: acc {:0.2f} bytes, per {:0.2f} bytes, veh {:0.2f} bytes'.format(accident.memory_usage(), person.memory_usage(deep=True), vehicle.memory_usage(deep=True)))
                print(accident.info())
                print(vehicle.info())
                print(person.info())
            del accident
            del person
            del vehicle

        if use_dask:
            people = dd.concat(lazy_people)
            vehicles = dd.concat(lazy_vehicles)
            accidents = dd.concat(lazy_accidents)
            people = people.compute()
            vehicles = vehicles.compute()
            accidents = accidents.compute()
        else:
            people = pd.concat(lazy_people)
            vehicles = pd.concat(lazy_vehicles)
            accidents = pd.concat(lazy_accidents)

        if debug > 0:
            print("Data loaded.")
            print("Processing People.")

        per = process_people(people, mappers)

        if debug > 0:
            print("Processing Accidents.")

        acc = process_accidents(accidents, mappers)

        if debug > 0:
            print("Processing Vehicles.")

        veh = process_vehicles(vehicles, mappers)

        save_pkl(target_folder, per, veh, acc)
    else:

        load_path = Path(__file__).resolve().parents[2] / "data" / "processed" / load_from

        veh = pd.read_pickle(load_path / "vehicles.pkl.xz")
        acc = pd.read_pickle(load_path / "accidents.pkl.xz")
        per = pd.read_pickle(load_path / "people.pkl.xz")

        if start_year > 1975 or end_year < 2018:
            veh = veh.query(f"YEAR >= {start_year} and YEAR <= {end_year}").reset_index()
            acc = veh.query(f"YEAR >= {start_year} and YEAR <= {end_year}").reset_index()
            per = per.query(f"YEAR >= {start_year} and YEAR <= {end_year}").reset_index()

    if debug > 0: print("Done")

    return veh, acc, per


def process_accidents(accidents, mappers):
    """Accident file processor, using predefined mapping dictionary.

    Accepts the unprocessed data from the concatenated ACCIDENTS.csv files. Applies appropriate mappings
    based on the year of each record, and returns a fully preprocessed DataFrame. Several specific mappings
    are applied from `fars_cleaner.extra_info`.

    Parameters
    ----------
    accidents : pandas.DataFrame
        Unprocessed DataFrame with accident-level data
    mappers : dict
        Dictionary containing mapping structure of selected years for all relevant variables in the accident table

    Returns
    -------
    pandas.DataFrame
        Processed accidents DataFrame
    """
    func_coalesce = []
    if 'ROAD_FNC' in accidents.columns:
        func_coalesce.append('ROAD_FNC')
    if 'FUNC_SYS' in accidents.columns:
        func_coalesce.append('FUNC_SYS')
    if len(func_coalesce) == 2:
        acc = accidents.coalesce(*func_coalesce, target_column_name='FUNCTION')
    elif len(func_coalesce) == 1:
        acc = accidents.rename({func_coalesce[0]: 'FUNCTION'})
    else:
        acc = accidents
    if 'FUNCTION' in acc.columns:
        acc = (acc
               .assign(FUNCTIONAL_CLASS=lambda x: ei.functional_class(x),
                       RURAL_OR_URBAN=lambda x: ei.land_use(x),
                       INTERSTATE=lambda x: ei.interstate(x), ))
    acc = (
        acc
            .assign(TIME_OF_DAY=lambda x: ei.time_of_day(x),
                    DAY_OF_WEEK=lambda x: ei.day_of_week(x),
                    COLLISION_TYPE=lambda x: ei.collision_type(x),
                    TRAFFICWAY=lambda x: ei.trafficway(x),
                    FIPS=lambda x: ei.get_fips(x),
                    )
            .groupby(['YEAR'])
            .apply(mapping, mappers=mappers['Accident'])
            .find_replace(match='exact',
                          CF1={45: pd.NA,
                               46: pd.NA,
                               60: pd.NA},
                          CF2={45: pd.NA,
                               46: pd.NA,
                               60: pd.NA},
                          CF3={45: pd.NA,
                               46: pd.NA,
                               60: pd.NA},

                          )
            # .droplevel(0)
            .remove_columns(column_names=['COUNTY'])
            .remove_empty()
    )
    if 'latitude' in acc.columns:
        acc = (
            acc
                .coalesce('LATITUDE', 'latitude')
                .coalesce('LONGITUD', 'longitud')
                .remove_columns(column_names=[
                                "latitude",
                                "longitud"]
            )
        )

    return acc


def fix_mod_year(df):
    df['MOD_YEAR'] = df['MOD_YEAR'].mask((df['MOD_YEAR'] < 99),
                                         df['MOD_YEAR'].add(1900))
    return df


def process_vehicles(vehicles, mappers):
    """Vehicle file processor, using predefined mapping dictionary.

    Accepts the unprocessed data from the concatenated VEHICLES.csv files. Applies appropriate mappings
    based on the year of each record, and returns a fully preprocessed DataFrame. Several specific mappings
    are applied from `fars_cleaner.extra_info`.

    Parameters
    ----------
    vehicles : pandas.DataFrame
        Unprocessed DataFrame with vehicle-level data
    mappers : dict
        Dictionary containing mapping structure of selected years for all relevant variables in the vehicle table

    Returns
    -------
    pandas.DataFrame
        Processed vehicles DataFrame
    """
    vehicles.reset_index(drop=True, inplace=True)

    veh = (
        vehicles
            .update_where(
            conditions=(vehicles['YEAR'] <= 1997) & (vehicles['MOD_YEAR'] == 99),
            target_column_name='MOD_YEAR',
            target_val=9999)
            .then(fix_mod_year)
            .assign(PASSENGER_CAR=lambda x: ei.is_passenger_car(x),
                    LIGHT_TRUCK_OR_VAN=lambda x:
                    ei.is_light_truck_or_van(x),
                    LARGE_TRUCK=lambda x: ei.is_large_truck(x),
                    MOTORCYCLE=lambda x: ei.is_motorcycle(x),
                    BUS=lambda x: ei.is_bus(x),
                    OTHER_UNKNOWN_VEHICLE=lambda x:
                    ei.is_other_or_unknown(x),
                    PASSENGER_VEHICLE=lambda x:
                    ei.is_passenger_vehicle(x),
                    UTILITY_VEHICLE=lambda x:
                    ei.is_utility_vehicle(x),
                    PICKUP=lambda x: ei.is_pickup(x),
                    VAN=lambda x: ei.is_van(x),
                    MEDIUM_TRUCK=lambda x: ei.is_medium_truck(x),
                    HEAVY_TRUCK=lambda x: ei.is_heavy_truck(x),
                    COMBINATION_TRUCK=lambda x:
                    ei.is_combination_truck(x),
                    SINGLE_UNIT_TRUCK=lambda x:
                    ei.is_single_unit_truck(x))
            .groupby(['YEAR'])
            .apply(mapping, mappers=mappers['Vehicle'])
            # .droplevel(0)
            .remove_empty()
    )

    if ('VEH_SC1' in veh.columns) and ('VEH_CF1' in veh.columns):
        veh = (
            veh
                # .coalesce(['VEH_CF1', 'VEH_SC1'])
                .coalesce('VEH_SC1', 'VEH_CF1', target_column_name='VEH_CF1')
                .coalesce('VEH_SC2', 'VEH_CF2', target_column_name='VEH_CF2')
                .coalesce('DR_SF1', 'DR_CF1', target_column_name='DR_CF1')
                .coalesce('DR_SF2', 'DR_CF2', target_column_name='DR_CF2')
                .coalesce('DR_SF3', 'DR_CF3', target_column_name='DR_CF3')
                .coalesce('DR_SF4', 'DR_CF4', target_column_name='DR_CF4')
                .remove_columns(column_names=[
                "VEH_SC1", "VEH_SC2",
                "DR_SF1", "DR_SF2", "DR_SF3", "DR_SF4",
            ]
            )
        )
    elif ('VEH_CF1' not in veh.columns):
        veh = (
            veh
                .rename_columns({
                'VEH_SC1': 'VEH_CF1',
                'VEH_SC2': 'VEH_CF2',
                'DR_SF1': 'DR_CF1',
                'DR_SF2': 'DR_CF2',
                'DR_SF3': 'DR_CF3',
                'DR_SF4': 'DR_CF4',
            }))

    return veh.encode_categorical(list(
        {'STATE', 'HIT_RUN', 'MAKE', 'BODY_TYP', 'ROLLOVER', 'J_KNIFE', 'TOW_VEH', 'SPEC_USE', 'EMER_USE', 'IMPACT1',
         'IMPACT2', 'DEFORMED', 'IMPACTS', 'TOWED', 'FIRE_EXP', 'VEH_CF1', 'VEH_CF2', 'M_HARM', 'WGTCD_TR', 'FUELCODE',
         'DR_PRES', 'DR_DRINK', 'L_STATE', 'L_STATUS', 'L_RESTRI', 'DR_TRAIN', 'VIOL_CHG', 'DR_CF1', 'DR_CF2', 'DR_CF3',
         'VINA_MOD', 'HAZ_CARG', 'VEH_MAN', 'L_COMPL', 'VIN_BT'} & set(veh.columns)))


def process_people(people, mappers):
    """Person file processor, using predefined mapping dictionary.

    Accepts the unprocessed data from the concatenated PERSON.csv files. Applies appropriate mappings
    based on the year of each record, and returns a fully preprocessed DataFrame. Several specific mappings
    are applied from `fars_cleaner.extra_info`.

    Parameters
    ----------
    people : pandas.DataFrame
        Unprocessed DataFrame with person-level data
    mappers : dict
        Dictionary containing mapping structure of selected years for all relevant variables in the person table

    Returns
    -------
    pandas.DataFrame
        Processed people DataFrame
    """
    if ('MAN_REST' in people.columns) and ('REST_USE' in people.columns):
        people = (
            people.coalesce('MAN_REST', 'REST_USE', target_column_name='REST_USE')
            .remove_columns(column_names=['MAN_REST'])
        )
    elif ('MAN_REST' in people.columns):
        people = (
            people.rename_columns({'MAN_REST':'REST_USE'})
        )

    per = (
        people
            .remove_empty()
            .assign(PERSON_TYPE=lambda x: ei.person_type(x),
                    RESTRAINTS=lambda x: ei.restraint_use(x),
                    AIR_BAG_DEPLOYMENT=lambda x: ei.air_bag_deployed(x),
                    DEAD=lambda x: x['INJ_SEV'] == 4,
                    HELMETED=lambda x: ei.helmet_use(x),
                    )
            .groupby(['YEAR'])
            .apply(mapping, mappers=mappers['Person'])
        # .droplevel(0)
    )

    if ('P_SF1' in per.columns) and ('P_CF1' in per.columns):
        per = (
            per
                .coalesce('P_SF1', 'P_CF1', target_column_name='P_CF1')
                .coalesce('P_SF2', 'P_CF2', target_column_name='P_CF2')
                .coalesce('P_SF3', 'P_CF3', target_column_name='P_CF3')
                .remove_columns(column_names=
                                ["P_SF1", "P_SF2", "P_SF3"]
                                )
        )
    elif 'P_CF1' not in per.columns:
        per = (
            per
                .rename_columns({
                'P_SF1': 'P_CF1',
                'P_SF2': 'P_CF2',
                'P_SF3': 'P_CF3',
            }))
    per = per.encode_categorical(list(
        {'STATE', 'SEX', 'PER_TYP', 'SEAT_POS', 'P_CF1', 'P_CF2', 'P_CF3'} & set(per.columns)))

    return createPerID(per, None)


def save_pkl(fname, people, vehicles, accidents):
    pkl_path = Path(__file__).resolve().parents[2] / "data" / "processed" / fname
    pkl_path.mkdir(parents=True, exist_ok=True)
    print("Saving accidents.pkl.xz ...")
    accidents.to_pickle(pkl_path / "accidents.pkl.xz")
    print("Saving people.pkl.xz ...")
    people.to_pickle(pkl_path / "people.pkl.xz")
    print('Saving vehicles.pkl.xz ...')
    vehicles.to_pickle(pkl_path / "vehicles.pkl.xz")


def decode_categorical(df):
    new_df = df.copy()
    for col in df.columns:
        if df[col].dtype.name == 'category':
            new_df[col] = new_df[col].astype(str)
    return new_df


def fix_mod_year(df):
    df['MOD_YEAR'] = df['MOD_YEAR'].mask((df['MOD_YEAR'] < 99),
                                         df['MOD_YEAR'].add(1900))
    return df


def mapping(group, mappers):
    yr = group.name
    cur_mappers = year_mapper(mappers, yr)
    if 'CARBUR' in cur_mappers.keys():
        cur_mappers['CARBUR'] = {str(k): v for k, v in cur_mappers['CARBUR'].items()}
        cur_mappers['FUELCODE'] = {str(k): v for k, v in cur_mappers['FUELCODE'].items()}
        cur_mappers['CYLINDER'] = {str(k): v for k, v in cur_mappers['CYLINDER'].items()}
        if 2011 <= yr <= 2012:
            cur_mappers['MCYCL_CY'] = {2: 'Two-Stroke', 4: 'Four-Stroke'}
            # {k:v for k,v in cur_mappers['MCYCL_CY'].items()}
        cur_mappers['TON_RAT'] = {str(k): v for k, v in cur_mappers['TON_RAT'].items()}
        cur_mappers['VIN_REST'] = {str(k): v for k, v in cur_mappers['VIN_REST'].items()}
        cur_mappers['VIN_BT'] = {str(k): v for k, v in cur_mappers['VIN_BT'].items()}
        cur_mappers['VINTYPE'] = {str(k): v for k, v in cur_mappers['VINTYPE'].items()}

    return group.replace(cur_mappers)


def load_basic(year, use_dask=False, data_dir=None, mapping=None, client=None):
    """Basic load pipeline to load FARS data for `year`

    Assumes FARS data for a given year is already downloaded and unzipped in `data_dir`, and
    requires a mapping dictionary to appropriately rename fields within the relevant dataframes.


    Parameters
    ----------
    year : int
        Year of data to load
    data_dir : Path-like
        Path where the data is stored. Must contain the folder "YYYY.unzip", where YYYY=`year`
    use_dask : bool, optional
        Flag to determine whether to parallelize using Dask. Supported, but not recommended at this time.
        The default is False.
    client : Dask Distributed Client, optional. Required if use_dask is True.
    mapping: dict
        mapping dictionary, as generated by fars_cleaner.builder.load_sheets()

    Returns
    -------
    (pandas.DataFrame, pandas.DataFrame, pandas.DataFrame)
        Returns three DataFrames: vehicles, accidents, and people.

    Raises
    ------
    ValueError
        If `use_dask` is True and no client provided.

    """

    if use_dask and client is None:
        raise ValueError("Must provide a client to use Dask.")

    cur_year = data_dir / f"{year}.unzip"


    cur_dir_files = os.listdir(cur_year)
    #veh_suggestions = fuzzyfinder("VEHICLE.csv", cur_dir_files)
    #per_suggestions = fuzzyfinder("PERSON.csv", cur_dir_files)
    #acc_suggestions = fuzzyfinder("ACCIDENT.csv", cur_dir_files)
    veh_suggestions = fuzzyprocess.extractOne("VEHICLE.csv", cur_dir_files)
    per_suggestions = fuzzyprocess.extractOne("PERSON.csv", cur_dir_files)
    acc_suggestions = fuzzyprocess.extractOne("ACCIDENT.csv", cur_dir_files)
    vehicle_fname = veh_suggestions[0]
    person_fname = per_suggestions[0]
    accident_fname = acc_suggestions[0]
    vehicle_file = cur_year / vehicle_fname
    person_file = cur_year / person_fname
    accident_file = cur_year / accident_fname
    print(vehicle_fname)

    acc_cols = get_renaming(mapping['Accident'], year)
    per_cols = get_renaming(mapping['Person'], year)
    veh_cols = get_renaming(mapping['Vehicle'], year)
    skip_veh = ['VE_FORMS', 'COUNTY', 'MONTH', 'DAY', 'HOUR', 'MINUTE',
                'ROAD_FNC', 'HARM_EV', 'MAN_COLL', 'SCH_BUS']
    skip_per = skip_veh + ['SCH_BUS', 'MAKE', 'BODY_TYP', 'MOD_YEAR', 'ROLLOVER',
                           'TOW_VEH', 'SPEC_USE', 'EMER_USE', 'IMPACT1', 'IMPACT2',
                           'IMPACTS', 'FIRE_EXP', 'WGTCD_TR', 'MAK_MOD', 'VIN_WGT',
                           'WHLBS_SH', 'WHLBS_LG', 'MCYCL_DS', 'VINA_MOD', 'SER_TR',
                           'VIN_BT', 'CERT_NO', 'VINTYPE', 'VINMAKE', 'VINMODYR',
                           'VIN_LNGT', 'FUELCODE', 'CARBUR', 'CYLINDER', 'DISPLACE',
                           'MCYCL_CY', 'TIRE_SZE', 'TON_RAT', 'TRK_WT', 'TRKWTVAR',
                           'MCYCL_WT', 'VIN_REST', 'WHLDRWHL', 'RUR_URB', 'FUNC_SYS',
                           'VPICMAKE', 'VPICMODEL', 'VPICBODYCLASS', 'ICFINALBODY']
    namedskip = []
    for skipper in skip_per:
        toap = f"{skipper}NAME"
        if toap not in skip_per:
            namedskip.append(toap)
    skip_per.extend(namedskip)
    namedskip = []
    for skipper in skip_veh:
        toap = f"{skipper}NAME"
        if toap not in skip_veh:
            namedskip.append(toap)
    skip_veh.extend(namedskip)
    if use_dask:

        veh_df = dd.from_pandas(pd.read_csv(vehicle_file,
                                            # encoding='ansi',
                                            encoding='cp1252',
                                            low_memory=False,
                                            usecols=lambda x: x not in skip_veh,
                                            dtype={'DEATHS': 'float64',
                                                   'OCUPANTS': 'float64'}),
                                npartitions=1).rename(columns=veh_cols)
        per_df = dd.from_pandas(pd.read_csv(person_file, encoding='cp1252',
                                            usecols=lambda x: x not in skip_per,
                                            low_memory=False),
                                npartitions=1).rename(columns=per_cols)

        acc_df = dd.from_pandas(pd.read_csv(accident_file, encoding='cp1252',
                                            low_memory=False),
                                npartitions=1).rename(columns=acc_cols)

        acc_df = dd.read_csv(accident_file, encoding='cp1252',
                             low_memory=False,
                             assume_missing=True).rename(columns=acc_cols)
    else:
        veh_df = pd.read_csv(vehicle_file,
                             # encoding='ansi',
                             encoding='cp1252',
                             low_memory=True,
                             usecols=lambda x: x not in skip_veh,
                             dtype={'DEATHS': 'float64',
                                    'OCUPANTS': 'float64',
                                    'MCARR_I1': 'str', 'MCARR_I2': 'str', 'MCARR_ID': 'str',
                                    'VIN_1': 'str', 'VIN_2': 'str', 'VIN_3': 'str',
                                    'VIN_4': 'str', 'VIN_5': 'str', 'VIN_5': 'str',
                                    'VIN_7': 'str', 'VIN_8': 'str', 'VIN_9': 'str',
                                    'VIN_10': 'str', 'VIN_11': 'str', 'VIN_12': 'str',
                                    'CYLINDER': 'str', 'TRLR1VIN': 'str', 'TRLR2VIN': 'str',
                                    'TRLR3VIN': 'str'
                                    }
                             ).rename(columns=veh_cols)
        per_df = pd.read_csv(person_file, encoding='cp1252',
                             usecols=lambda x: x not in skip_per and not x.endswith("NAME"),
                             low_memory=True).rename(columns=per_cols)
        acc_df = pd.read_csv(accident_file, encoding='cp1252',
                             usecols=lambda x: not x.endswith("NAME"),
                             low_memory=True).rename(columns=acc_cols)

    if not use_dask:
        acc_intCols = acc_df.select_dtypes('integer').columns
        acc_floatCols = acc_df.select_dtypes('float').columns
        acc_df[acc_intCols] = acc_df[acc_intCols].apply(pd.to_numeric, downcast='integer')
        acc_df[acc_floatCols] = acc_df[acc_floatCols].apply(pd.to_numeric, downcast='float')

        veh_intCols = veh_df.select_dtypes('integer').columns
        veh_floatCols = veh_df.select_dtypes('float').columns
        veh_df[veh_intCols] = veh_df[veh_intCols].apply(pd.to_numeric, downcast='integer')
        veh_df[veh_floatCols] = veh_df[veh_floatCols].apply(pd.to_numeric, downcast='float')

        per_intCols = per_df.select_dtypes('integer').columns
        per_floatCols = per_df.select_dtypes('float').columns
        per_df[per_intCols] = per_df[per_intCols].apply(pd.to_numeric, downcast='integer')
        per_df[per_floatCols] = per_df[per_floatCols].apply(pd.to_numeric, downcast='float')
	
    return veh_df, per_df, acc_df



# -*- coding: utf-8 -*-
"""
Created on Mon Feb 17 11:04:59 2020

@author: Mitchell Abrams
"""


import pandas as pd
import janitor
import datetime

import pickle

from pathlib import Path

#from builder import *
from fars_cleaner.builder import get_renaming
import fars_cleaner.extra_info as ei

from fars_cleaner.fars_utils import createPerID

from fars_cleaner import FARSFetcher

class FARSProcessor:
    def __init__(self,
                 start_year=1975,
                 end_year=2018,
                 fetcher: FARSFetcher = None,
                 ):
        """


        Parameters
        ----------
        start_year : int, optional
            Year to start analysis. The default is 1975.
        end_year : int, optional
            Year to end analysis. The default is 2018.
        first_run : bool, optional
            Flag to determine whether to process and write-out the required files, or whether the data can be loaded from
            disk pre-processed. The default is True.
        use_dask : bool, optional
            Flag to determine whether to parallelize using Dask. Not implemented at this time.
            The default is False.
        client : Dask Distributed Client, optional. Required if use_dask is True. Not implemented at this time.

        Returns
        -------


        """
        self.NOW = datetime.datetime.now()
        self.start_year = start_year
        self.end_year = end_year

        if fetcher is None:
            self.fetcher = FARSFetcher()
        else:
            self.fetcher = fetcher
            
        fetcher.fetch_subset(self.start_year, self.end_year)

        self.data_dir = self.fetcher.get_data_path()
        with open(fetcher.fetch_mappers(), 'rb') as f:
            self.mappers = pickle.load(f)
        self.load_paths = self.fetcher.fetch_subset(self.start_year, self.end_year)

        people = []
        vehicles = []
        accidents = []

        for year in range(start_year, end_year + 1):
            vehicle = self.load_vehicles(year)
            person = self.load_people(year)
            accident = self.load_accidents(year)

            accident['YEAR'] = year
            vehicle['YEAR'] = year
            person['YEAR'] = year

            people.append(person)
            vehicles.append(vehicle)
            accidents.append(accident)
        #self.accidents = pd.concat(accidents)
        print("accidents")
        self.accidents = self.process_accidents(pd.concat(accidents))
        print("people")
        self.people = self.process_people(pd.concat(people))
        print("vehicles")
        self.vehicles = self.process_vehicles(pd.concat(vehicles))

    def year_mapper(self, mappers, year):
        # Take mapper for the data file
        # Iterate through: for each sub-dictionary that was implemented before
        # and discontinued after this year
        cur_mapper = {code: {} for code in mappers.keys()}
        for code, detail in mappers.items():
            discon = detail['discontinued']
            if not discon:
                discon = self.NOW.year

            if detail['implemented'] <= year <= discon:
                # CONDITIONAL FOR RENAMING VARIABLES THAT CHANGED NAMES
                if detail['df_name']:
                    code = detail['df_name']
                if detail['lookup']:
                    cur_mapper[code] = detail['mappers'][year]

        return cur_mapper

    def mapping(self, group, mappers):
        yr = group.name
        cur_mappers = self.year_mapper(mappers, yr)
        if 'CARBUR' in cur_mappers.keys():
            cur_mappers['CARBUR'] = {str(k): v for k, v in cur_mappers['CARBUR'].items()}
            cur_mappers['FUELCODE'] = {str(k): v for k, v in cur_mappers['FUELCODE'].items()}
            cur_mappers['CYLINDER'] = {str(k): v for k, v in cur_mappers['CYLINDER'].items()}
            if 2011 <= yr <= 2012:
                cur_mappers['MCYCL_CY'] = {2: 'Two-Stroke', 4: 'Four-Stroke'}
                # {k:v for k,v in cur_mappers['MCYCL_CY'].items()}
            cur_mappers['TON_RAT'] = {str(k): v for k, v in cur_mappers['TON_RAT'].items()}
            cur_mappers['VIN_REST'] = {str(k): v for k, v in cur_mappers['VIN_REST'].items()}
            cur_mappers['VIN_BT'] = {str(k): v for k, v in cur_mappers['VIN_BT'].items()}
            cur_mappers['VINTYPE'] = {str(k): v for k, v in cur_mappers['VINTYPE'].items()}

        return group.replace(cur_mappers)

    def process_accidents(self, accidents):

        acc = (
            accidents
                .assign(TIME_OF_DAY=lambda x: ei.time_of_day(x),
                        DAY_OF_WEEK=lambda x: ei.day_of_week(x),
                        COLLISION_TYPE=lambda x: ei.collision_type(x),
                        TRAFFICWAY=lambda x: ei.trafficway(x),
                        FUNCTIONAL_CLASS=lambda x: ei.functional_class(x),
                        RURAL_OR_URBAN=lambda x: ei.land_use(x),
                        INTERSTATE=lambda x: ei.interstate(x),
                        FIPS=lambda x: ei.get_fips(x),
                        )
                .groupby(['YEAR'])
                .apply(self.mapping, mappers=self.mappers['Accident'])
                .find_replace(match='exact',
                              CF1={45: pd.NA,
                                   46: pd.NA,
                                   60: pd.NA},
                              CF2={45: pd.NA,
                                   46: pd.NA,
                                   60: pd.NA},
                              CF3={45: pd.NA,
                                   46: pd.NA,
                                   60: pd.NA},

                              )
                .droplevel(0)
                .remove_empty()
        )
        if 'latitude' in acc.columns:
            acc = (
                acc
                .coalesce(['LATITUDE', 'latitude'])
                .coalesce(['LONGITUD', 'longitud'])
            )

        return acc

    def fix_mod_year(self, df):
        df['MOD_YEAR'] = df['MOD_YEAR'].mask((df['MOD_YEAR'] < 99),
                                             df['MOD_YEAR'].add(1900))
        return df

    def process_vehicles(self, vehicles):
        vehicles.reset_index(drop=True, inplace=True)

        veh = (
            vehicles
                .update_where(
                conditions=(vehicles['YEAR'] <= 1997) & (vehicles['MOD_YEAR'] == 99),
                target_column_name='MOD_YEAR',
                target_val=9999)
                .then(self.fix_mod_year)
                .assign(PASSENGER_CAR=lambda x: ei.is_passenger_car(x),
                        LIGHT_TRUCK_OR_VAN=lambda x:
                        ei.is_light_truck_or_van(x),
                        LARGE_TRUCK=lambda x: ei.is_large_truck(x),
                        MOTORCYCLE=lambda x: ei.is_motorcycle(x),
                        BUS=lambda x: ei.is_bus(x),
                        OTHER_UNKNOWN_VEHICLE=lambda x:
                        ei.is_other_or_unknown(x),
                        PASSENGER_VEHICLE=lambda x:
                        ei.is_passenger_vehicle(x),
                        UTILITY_VEHICLE=lambda x:
                        ei.is_utility_vehicle(x),
                        PICKUP=lambda x: ei.is_pickup(x),
                        VAN=lambda x: ei.is_van(x),
                        MEDIUM_TRUCK=lambda x: ei.is_medium_truck(x),
                        HEAVY_TRUCK=lambda x: ei.is_heavy_truck(x),
                        COMBINATION_TRUCK=lambda x:
                        ei.is_combination_truck(x),
                        SINGLE_UNIT_TRUCK=lambda x:
                        ei.is_single_unit_truck(x))
                .groupby(['YEAR'])
                .apply(self.mapping, mappers=self.mappers['Vehicle'])
                .droplevel(0)
                .remove_empty()
        )

        if ('VEH_SC1' in veh.columns) and ('VEH_CF1' in veh.columns):
            veh = (
                veh
                    .coalesce(['VEH_CF1', 'VEH_SC1'])
                    .coalesce(['VEH_CF2', 'VEH_SC2'])
                    .coalesce(['DR_CF1', 'DR_SF1'])
                    .coalesce(['DR_CF2', 'DR_SF2'])
                    .coalesce(['DR_CF3', 'DR_SF3'])
                    .coalesce(['DR_CF4', 'DR_SF4'])
            )
        elif ('VEH_CF1' not in veh.columns):
            veh = (
                veh
                .rename_columns({
                    'VEH_SC1': 'VEH_CF1',
                    'VEH_SC2': 'VEH_CF2',
                    'DR_SF1': 'DR_CF1',
                    'DR_SF2': 'DR_CF2',
                    'DR_SF3': 'DR_CF3',
                    'DR_SF4': 'DR_CF4',
                }))

        return veh.encode_categorical(list(set(['STATE', 'HIT_RUN', 'MAKE', 'BODY_TYP',
                                     'ROLLOVER', 'J_KNIFE', 'TOW_VEH', 'SPEC_USE', 'EMER_USE', 'IMPACT1',
                                     'IMPACT2', 'DEFORMED', 'IMPACTS', 'TOWED', 'FIRE_EXP', 'VEH_CF1',
                                     'VEH_CF2', 'M_HARM', 'WGTCD_TR', 'FUELCODE', 'DR_PRES', 'DR_DRINK',
                                     'L_STATE', 'L_STATUS', 'L_RESTRI', 'DR_TRAIN', 'VIOL_CHG',
                                     'DR_CF1', 'DR_CF2', 'DR_CF3', 'VINA_MOD', 'HAZ_CARG', 'VEH_MAN',
                                     'L_COMPL', 'VIN_BT']) & set(veh.columns)))

    def process_people(self, people):
        per = (
            people
                .remove_empty()
                .assign(PERSON_TYPE=lambda x: ei.person_type(x),
                        RESTRAINTS=lambda x: ei.restraint_use(x),
                        AIR_BAG_DEPLOYMENT=lambda x: ei.air_bag_deployed(x),
                        DEAD=lambda x: x['INJ_SEV'] == 4,
                        HELMETED=lambda x: ei.helmet_use(x),
                        )
                .groupby(['YEAR'])
                .apply(self.mapping, mappers=self.mappers['Person'])
                .droplevel(0)
        )
        if ('P_SF1' in per.columns) and ('P_CF1' in per.columns):
            per = (
                per
                .coalesce(['P_CF1', 'P_SF1'])
                .coalesce(['P_CF2', 'P_SF2'])
                .coalesce(['P_CF3', 'P_SF3'])
            )
        elif ('P_CF1' not in per.columns):
            per = (
                per
                .rename_columns({
                    'P_SF1': 'P_CF1',
                    'P_SF2': 'P_CF2',
                    'P_SF3': 'P_CF3',
                }))

        return createPerID(per, None)

    def get_renaming(self, map_with):
        """Get original to final column namings."""
        renamers = {}
        for code, attr in self.mappers[map_with].items():
            renamers[code] = attr['df_name']
        return renamers

    def load_accidents(self, year):
        accident_file = self.data_dir / f"{year}.unzip" / "ACCIDENT.csv"
        acc_cols = self.get_renaming('Accident')

        acc_df = pd.read_csv(accident_file, encoding='cp1252',
                             low_memory=False).rename(columns=acc_cols)

        return acc_df

    def load_vehicles(self, year):
        vehicle_file = self.data_dir / f"{year}.unzip" / "VEHICLE.csv"
        veh_cols = self.get_renaming('Vehicle')
        skip_cols = ['VE_FORMS', 'COUNTY', 'MONTH', 'DAY', 'HOUR', 'MINUTE',
                     'ROAD_FNC', 'HARM_EV', 'MAN_COLL', 'SCH_BUS']

        veh_df = pd.read_csv(vehicle_file,
                             # encoding='ansi',
                             encoding='cp1252',
                             low_memory=False,
                             usecols=lambda x: x not in skip_cols,
                             dtype={'DEATHS': 'float64',
                                    'OCUPANTS': 'float64'}
                             ).rename(columns=veh_cols)

        return veh_df

    def load_people(self, year):
        person_file = self.data_dir / f"{year}.unzip" / "PERSON.csv"
        per_cols = self.get_renaming('Person')
        skip_cols = ['VE_FORMS', 'COUNTY', 'MONTH', 'DAY', 'HOUR', 'MINUTE',
                     'ROAD_FNC', 'HARM_EV', 'MAN_COLL', 'SCH_BUS', 'SCH_BUS',
                     'MAKE', 'BODY_TYP', 'MOD_YEAR', 'ROLLOVER',
                     'TOW_VEH', 'SPEC_USE', 'EMER_USE', 'IMPACT1', 'IMPACT2',
                     'IMPACTS', 'FIRE_EXP', 'WGTCD_TR', 'MAK_MOD', 'VIN_WGT',
                     'WHLBS_SH', 'WHLBS_LG', 'MCYCL_DS', 'VINA_MOD', 'SER_TR',
                     'VIN_BT', 'CERT_NO', 'VINTYPE', 'VINMAKE', 'VINMODYR',
                     'VIN_LNGT', 'FUELCODE', 'CARBUR', 'CYLINDER', 'DISPLACE',
                     'MCYCL_CY', 'TIRE_SZE', 'TON_RAT', 'TRK_WT', 'TRKWTVAR',
                     'MCYCL_WT', 'VIN_REST', 'WHLDRWHL', 'RUR_URB', 'FUNC_SYS']

        per_df = pd.read_csv(person_file, encoding='cp1252',
                             usecols=lambda x: x not in skip_cols,
                             low_memory=False).rename(columns=per_cols)
        return per_df


"""
Load sample data.
"""
import pooch
import os

from pathlib import Path

class FARSFetcher:
    def __init__(self,
                 cache_path=None,
                 registry=None,
                 project_dir=None,
                 check_hash=True,
                 show_progress=True,
                 ):
        """Class to download FARS data from the NHTSA FTP repository.

        Note that on first run, this will take a long time to fully download the data, as the repository is large.
        Expect first run to take 5-10+ minutes, depending on your setup.

        Parameters
        ----------
        cache_path: `os.path` or path-like, or str, optional
            The path to save the downloaded FARS files to.
            Default is `pooch.os_cache("fars")`, the default cache path as defined by the OS. See `pooch` and
            `appdirs` documentations.
            If `str`, and `project_dir` is not `None`, files will be downloaded to `project_dir/cache_path`
        registry:
            Path to registry file. Defaults to path for packaged `registry.txt` file. Override at your own risk.
        project_dir:
            Top level directory for your current project. If a path is provided, and `cache_path` is left as default,
            files will be downloaded to `project_dir/data/fars`. If `cache_path` is not the default, files will be
            downloaded to `project_dir/cache_path`.
        check_hash: bool
            Flag to enforce pooch download behavior. Defaults to True. When False, force download of FARS resources
            regardless of hash mismatch against the local registry version. Useful for when the FARS
            database is updated before the registry can be modified. Should normally be left to default (False).
        show_progress: bool
            Use pooch built-in feature to show progress bars during download. Default True.
        """
        if project_dir:
            self.project_dir = project_dir
            if cache_path:
                self.cache_path = Path(project_dir) / cache_path
            else:
                self.cache_path = Path(project_dir) / "data" / "fars"
            self.project_dir.mkdir(parents=True, exist_ok=True)
            self.cache_path.mkdir(parents=True, exist_ok=True)
        else:
            self.project_dir = None
            if cache_path:
                self.cache_path = Path(cache_path)
                self.cache_path.mkdir(parents=True, exist_ok=True)
            else:
                self.cache_path = pooch.os_cache("fars")

        if registry:
            self.registry = Path(registry)
        else:
            self.registry = os.path.join(os.path.dirname(__file__), "registry.txt")

        self.check_hash = check_hash
        self.show_progress = show_progress

        self.GOODBOY = pooch.create(
            path=self.cache_path,
            base_url="https://www.nhtsa.gov/filebrowser/download/",
            registry=None,
            allow_updates=self.check_hash,
        )

        self.GOODBOY.load_registry(self.registry)

    def fetch_all(self):
        """
        Download the entire FARS dataset, to cache folder.
        """
        # The file will be downloaded automatically the first time this is run.

        fnames = self.GOODBOY.registry_files
        unzipped = {}
        
        for fname in fnames:

            if self.GOODBOY.is_available(fname):
                if "dict" in fname:
                    self.GOODBOY.fetch(fname, progressbar=self.show_progress)
                else:
                    unpack = pooch.Unzip(extract_dir=f"./{fname[:-4]}.unzip")
                    unzipped[fname] = self.GOODBOY.fetch(fname, processor=unpack, progressbar=self.show_progress)
            else:
                raise FileNotFoundError("File could not be found in FARS FTP directory.")
        return unzipped

    def fetch_subset(self, start_yr, end_yr):
        """
        Download a subset of the FARS dataset.
        """
        unzipped = {}
        for yr in range(start_yr, end_yr + 1):
            unzipped[yr] = self.fetch_single(yr)
        return unzipped

    def fetch_single(self, year):
        """
        Load the FARS data for a given year.
        """
        # fname = f'{year}/National/FARS{year}NationalCSV.zip'
        fname = f'{year}.zip'
        if self.GOODBOY.is_available(fname):
            unpack = pooch.Unzip(extract_dir=f"./{fname[:-4]}.unzip")
            unzipped = self.GOODBOY.fetch(fname, processor=unpack, progressbar=self.show_progress)
        else:
            raise FileNotFoundError(f"{fname}: File could not be found in FARS FTP directory.")

        return {year: unzipped}

    def fetch_mappers(self):
        """
        Loads the mappings for each variable from a pickled dictionary.

        Returns
        -------
        Path to the mapper file
        """
        return self.GOODBOY.fetch("mapping.dict")

    def get_data_path(self):
        return self.cache_path
    
    def get_show_progress(self):
        return self.show_progress



# -*- coding: utf-8 -*-
"""
Created on Mon Jan 27 10:07:15 2020

@author: Mitchell Abrams

"""
import datetime

import pandas as pd
#import extra_info as ei

from pathlib import Path
import pickle

from time import sleep

NOW = datetime.datetime.now()

def get_renaming(mappers, year):
    """Get original to final column namings."""
    renamers = {}
    for code, attr in mappers.items():
        renamers[code] = attr['df_name']
    return renamers


def year_mapper(mappers, year):
    # Take mapper for the data file
    # Iterate through: for each sub-dictionary that was implemented before
    # and discontinued after this year
    cur_mapper = {code: {} for code in mappers.keys()}
    for code, detail in mappers.items():
        discon = detail['discontinued']
        if not discon:
            discon = NOW.year

        if detail['implemented'] <= year <= discon:
            # CONDITIONAL FOR RENAMING VARIABLES THAT CHANGED NAMES
            if detail['df_name']:
                code = detail['df_name']
            if detail['lookup']:
                cur_mapper[code] = detail['mappers'][year]

    return cur_mapper


def load_sheets(t_list=None,
                table_folder=Path(__file__).parent.resolve() / "lookup_tables"):
    """Load mapping for given sheets.

    Loads a list of .xlsx files from disk as named in `t_list`. Generates a
    nested dict through which individual codes can be accessed and decoded to
    a human-readable format.

    Parameters
    ----------
    table_folder: Path-like
        Path to folder with lookup tables. Defaults to "lookup_tables"
        folder in current directory.
    t_list : list, optional
        List of code sheets to load and map. The default is:
            ['Accident', 'Vehicle', ' Person', 'Vehnit'].

    Returns
    -------
    mappers : multi-level nested dict
        `mappers` is a nested dictionary. Level hierarachy is:
            File_Name:
                Field_Name:
                    `description`: str
                        Short description of field.
                    `discontinued`: float64 or bool
                        Last year variable appears in data set.
                        False if not yet discontinued.
                    `implemented`: float64
                        First year variable appears in data set
                    `mappers`: dict
                        YEAR: dict (keys are int)
                            key:value pairs by year
                    `df_name`: str
                        Variable name to use in dataframe

    See Also
    --------
    lookup: Generate dictionary mapping for single code in a single year.

    """
    if t_list is None:
        t_list = ['Accident', 'Vehicle', 'Person', 'Vehnit']

    mappers = {}

    template = {'description': None,
                'rename': False,
                'implemented': 1975,
                'discontinued': False,
                'dtype': None,
                'lookup': False,
                'mappers': None,}
    #table_folder = Path(__file__).parent.resolve()
    for table in t_list:

        target_table = table_folder / f"{table}.xlsx"

        df = pd.read_excel(target_table, sheet_name=None)

        df['Summary']['Year_Discontinued'].fillna(int(NOW.year), inplace=True)
        df['Summary']['Years_Skipped'].fillna(-1, inplace=True)
        df['Summary'].set_index('Code', inplace=True)

        mapper = {code: template.copy()
                  for code in df['Summary'].index.values}

        summary = df['Summary']

        for code in mapper.keys():

            impl = summary.at[code, 'Year_Implemented']
            discon = summary.at[code, 'Year_Discontinued']
            df_name = summary.at[code, 'Rename']

            mapper[code] = {'description': summary.at[code, 'Description'],
                            'use_name': summary.at[code, 'Use_Sheet'],
                            'implemented': impl,
                            'discontinued': discon if discon != 2020 else False,
                            'lookup': bool(summary.at[code, 'Lookup']),
                            'df_name': df_name if not pd.isna(df_name) else code
                            }
            lookup_me = mapper[code]['lookup']

            if lookup_me:
                mapper[code]['mappers'] = {yr: {} for
                                           yr in range(impl, int(discon)+1)}
                if not pd.isna(summary.at[code, 'Coded_In']):
                    secondary_target = table_folder / "{}.xlsx".format(
                        summary.at[code, 'Coded_In'])
                    target_sheet = pd.read_excel(
                        secondary_target,
                        summary.at[code, 'Use_Sheet'])
                else:
                    target_sheet = df[mapper[code]['use_name']]
                #print(code)
                for year in range(impl, int(discon)+1):
                    if impl <= year <= discon:# and discon >= year:
                        mapper[code]['mappers'][year] = dict(
                            lookup(target_sheet, year))
            else:
                mapper[code]['mappers'] = None

        mappers[table] = mapper.copy()

    return mappers


def lookup(code_table, year):
    """Generate a dictionary mappping for the key-value pairing for one code.

    Filters a dataframe to contain only code mappings valid during the
    specified `year`. Returns the dictionary representation of these code
    mappings (code:decoded_value format). Typically, encoded values are int
    and decoded values are str.

    Parameters
    ----------
    code_table : pd.DataFrame
        Dataframe with all key-value mappings for all years for a single code.
    year : int
        The year to pull mappings for.

    Returns
    -------
    dict
        Returns a dictionary with the key:value mapping for `year`.

    See Also
    --------
    load_sheets: loads the data sheets and passes to this for mapping
    """
    cur_lookup = code_table.fillna({'Year_Implemented': 1975,
                                    'Year_Discontinued': 2020})
    cur_lookup = cur_lookup.query("Year_Implemented <= {year} and "
                                  "Year_Discontinued >= {year}".format(year=year))
    return pd.Series(cur_lookup['DEF'].values, index=cur_lookup['ID'].values).to_dict()


if __name__ == '__main__':

    m = load_sheets()
    # for year in range(1975, 2019):
    #     year_mapper(m['Accident'], year)
    #     year_mapper(m['Person'], year)
    #     year_mapper(m['Vehicle'], year)
    #     year_mapper(m['Vehnit'], year)




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fars_cleaner']

package_data = \
{'': ['*'], 'fars_cleaner': ['lookup_tables/*']}

install_requires = \
['dask',
 'distributed>=2022,<2023',
 'pathlib',
 'pooch>=1.6.0',
 'pyjanitor>=0.23.1,<0.24.0',
 'requests',
 'scipy>=1.7.0,<2.0.0',
 'thefuzz',
 'tqdm']

extras_require = \
{':python_version >= "3.8"': ['pandas>=1.4,<2.0', 'numpy>=1.22.0,<2.0.0'],
 'dev': ['pytest>=7.1.0,<8.0.0', 'hypothesis']}

setup_kwargs = {
    'name': 'fars-cleaner',
    'version': '1.3.1',
    'description': 'A package for loading and preprocessing the NHTSA FARS crash database',
    'long_description': '![GitHub release (latest SemVer)](https://img.shields.io/github/v/release/mzabrams/fars-cleaner)\n![PyPI](https://img.shields.io/pypi/v/fars-cleaner)\n[![License](https://img.shields.io/badge/License-BSD%203--Clause-blue.svg)](https://opensource.org/licenses/BSD-3-Clause)\n[![DOI](https://zenodo.org/badge/252038452.svg)](https://zenodo.org/badge/latestdoi/252038452)\n\n[![status](https://joss.theoj.org/papers/2ca54c6935611fe3cb0303c49a354c51/status.svg)](https://joss.theoj.org/papers/2ca54c6935611fe3cb0303c49a354c51)\n\n# FARS Cleaner `fars-cleaner`\n\n`fars-cleaner` is a Python library for downloading and pre-processing data \nfrom the Fatality Analysis Reporting System, collected annually by NHTSA since\n 1975. \n\n## Installation\n\nThe preferred installation method is through `conda`.\n```bash\nconda install -c conda-forge fars-cleaner\n```\nYou can also install with [pip](https://pip.pypa.io/en/stable/).\n\n```bash\npip install fars-cleaner\n```\n\n## Usage\n\n### Downloading FARS data\nThe `FARSFetcher` class provides an interface to download and unzip selected years from the NHTSA FARS FTP server. \nThe class uses `pooch` to download and unzip the selected files. By default, files are unzipped to your OS\'s cache directory.\n\n```python\nfrom fars_cleaner import FARSFetcher\n\n# Prepare for FARS file download, using the OS cache directory. \nfetcher = FARSFetcher()\n```\nSuggested usage is to download files to a data directory in your current project directory. \nPassing `project_dir` will download files to `project_dir/data/fars` by default. This behavior can be \noverridden by setting `cache_path` as well. Setting `cache_path` alone provides a direct path to the directory\nyou want to download files into.\n```python\nfrom pathlib import Path\nfrom fars_cleaner import FARSFetcher\n\nSOME_PATH = Path("/YOUR/PROJECT/PATH") \n# Prepare to download to /YOUR/PROJECT/PATH/data/fars\n# This is the recommended usage.\nfetcher = FARSFetcher(project_dir=SOME_PATH)\n\n# Prepare to download to /YOUR/PROJECT/PATH/fars\ncache_path = "fars"\nfetcher = FARSFetcher(project_dir=SOME_PATH, cache_path=cache_path)\n\ncache_path = Path("/SOME/TARGET/DIRECTORY")\n# Prepare to download directly to a specific directory.\nfetcher = FARSFetcher(cache_path=cache_path)\n```\n\nFiles can be downloaded in their entirety (data from 1975-2018), as a single year, or across a specified year range.\nDownloading all of the data can be quite time consuming. The download will simultaneously unzip the folders, and delete \nthe zip files. Each zipped file will be unzipped and saved in a folder `{YEAR}.unzip`\n```python\n# Fetch all data\nfetcher.fetch_all()\n\n# Fetch a single year\nfetcher.fetch_single(1984)\n\n# Fetch data in a year range (inclusive).\nfetcher.fetch_subset(1999, 2007)\n```\n\n### Processing FARS data\nCalling `load_pipeline` will allow for full loading and pre-processing of the FARS data requested by the user.\n```python\nfrom fars_cleaner import FARSFetcher, load_pipeline\n\nfetcher = FARSFetcher(project_dir=SOME_PATH)\nvehicles, accidents, people = load_pipeline(fetcher=fetcher,\n                                            first_run=True,\n                                            target_folder=SOME_PATH)\n```\n\nCalling `load_basic` allows for simple loading of the FARS data for a single year, with no preprocessing. Files must\nbe prefetched using a `FARSFetcher` or similar method. A `mapper` dictionary must be provided to identify what, if \nany, columns require renaming. \n\n```python\nfrom fars_cleaner.data_loader import load_basic\n\nvehicles, accidents, people = load_basic(year=1975, data_dir=SOME_PATH, mapping=mappings)\n```\n\n## Requirements\nDownloading and processing the full FARS dataset currently runs out of memory on Windows machines with only 16GB RAM. It is recommended to have at least 32GB RAM on Windows systems. macOS and Linux run with no issues on 16GB systems.\n\n## Contributing\nPull requests are welcome. For major changes, please open an issue first to discuss what you would like to change. See [CONTRIBUTING.md](CONTRIBUTING.md) for more details.\n\n## License\n[BSD-3 Clause](https://choosealicense.com/licenses/bsd-3-clause/)\n',
    'author': 'Mitchell Abrams',
    'author_email': 'mitchell.abrams@duke.edu',
    'maintainer': 'Mitchell Abrams',
    'maintainer_email': 'mitchell.abrams@duke.edu',
    'url': 'https://github.com/mzabrams/fars-cleaner',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



name = 'fars_cleaner'
__version__ = "1.2.0"

from fars_cleaner.datasets import FARSFetcher
from fars_cleaner.fars_processor import FARSProcessor
from fars_cleaner.data_loader import load_pipeline
#from .data_loader import pipeline
